
/* gcc-plugin.cc */

// http://codesynthesis.com/~boris/blog/2010/05/17/parsing-cxx-with-gcc-plugin-part-3/
// http://www.codesynthesis.com/~boris/data/gcc-plugin/plugin-3.cxx

// see <gcc-sources>/gcc/testsuite/g++.dg/plugin/decl_plugin.c

#include <gmp.h>

#include "gcc-plugin.h"
#include "plugin-version.h"


#include "cp/cp-tree.h"
// #include "c-common.h"
// #include "c-pragma.h"
// #include "diagnostic.h"

int plugin_is_GPL_compatible;

#if GCCPLUGIN_VERSION < 4009
inline const char * tree_name (int tc) { return tree_code_name [tc]; }
#else
inline const char * tree_name (int tc) { return get_tree_code_name ((tree_code) tc); }
#endif

#include <iostream>
#include "support.h"

using namespace NAMESPACE;
using namespace std;

TTextOutput io;

/* ---------------------------------------------------------------------- */

void print_decl (tree decl);

string decl_scope (tree decl)
{
   string s, tmp;

   for (tree scope = CP_DECL_CONTEXT (decl);
        scope != global_namespace && scope != NULL;
        scope = CP_DECL_CONTEXT (scope))
   {
      if (TREE_CODE (scope) == RECORD_TYPE)
         scope = TYPE_NAME (scope);

      tree id = DECL_NAME (scope);

      tmp = "::";
      tmp += (id != 0 ? IDENTIFIER_POINTER (id) : "<unnamed>");
      tmp += s;
      s.swap (tmp);
   }

   return s;
}

/* ---------------------------------------------------------------------- */

enum access_spec
{
   public_, protected_, private_
};

const char* access_spec_str[] =
{
   "public", "protected", "private"
};

/* ---------------------------------------------------------------------- */

void print_class (tree type)
{
   type = TYPE_MAIN_VARIANT (type);

   tree decl = TYPE_NAME (type);
   tree id  = DECL_NAME (decl);
   const char * name = IDENTIFIER_POINTER (id);

   io.Put ("class " + decl_scope (decl) + name);
   if (!COMPLETE_TYPE_P (type)) io.Put (";");

   io.PutLn (string (" // ") +
             DECL_SOURCE_FILE (decl) +  ":" +
             IntToStr (DECL_SOURCE_LINE (decl)) + ":" +
             IntToStr (DECL_SOURCE_COLUMN (decl)));

   // We are done if this is an incomplete
   // class declaration.
   //
   if (COMPLETE_TYPE_P (type))
   {
      io.PutLn ("{");
      io.IncIndent ();

      // Traverse base information.
      //
      tree biv = TYPE_BINFO (type);
      size_t n = biv ? BINFO_N_BASE_BINFOS (biv) : 0;

      for (size_t i = 0; i < n; i++)
      {
        tree bi = BINFO_BASE_BINFO (biv, i);

        // Get access specifier.
        //
        access_spec a = public_;

        if (BINFO_BASE_ACCESSES (biv))
        {
          tree ac = BINFO_BASE_ACCESS (biv, i);

         if (ac == NULL || ac == access_public_node)
             a = public_;
          else if (ac == access_protected_node)
             a = protected_;
          else
             a = private_;
        }

        bool virt = BINFO_VIRTUAL_P (bi);
        tree b_type = TYPE_MAIN_VARIANT (BINFO_TYPE (bi));
        tree b_decl = TYPE_NAME (b_type);
        tree b_id = DECL_NAME (b_decl);
        const char * b_name = IDENTIFIER_POINTER (b_id);

        // cerr << "\t" << access_spec_str[a] << (virt ? " virtual" : "")
        //      << " base " << decl_scope (b_decl) << "::" << b_name << endl;
      }

      // Traverse members.
      //
      for (tree d = TYPE_FIELDS (type); d != NULL; d = TREE_CHAIN (d))
      {
         switch (TREE_CODE (d))
         {
            case TYPE_DECL:
               if (!DECL_SELF_REFERENCE_P (d))
                  print_decl (d);
               break;

            case FIELD_DECL:
               if (!DECL_ARTIFICIAL (d))
                  print_decl (d);
               break;

            default:
               print_decl (d);
               break;
         }
      }

      for (tree d = TYPE_METHODS (type); d != NULL; d = TREE_CHAIN (d))
      {
        if (! DECL_ARTIFICIAL (d))
           print_decl (d);
      }

      io.DecIndent ();
      io.PutLn ("};");
   }
}

/* ---------------------------------------------------------------------- */

void print_decl (tree decl)
{
   int dc = TREE_CODE (decl);
   tree type = TREE_TYPE (decl);
   int tc = 0;

   if (type)
   {
      tc = TREE_CODE (type);

      if (dc == TYPE_DECL && tc == RECORD_TYPE)
      {
        // If DECL_ARTIFICIAL is true this is a class
        // declaration. Otherwise this is a typedef.
        if (DECL_ARTIFICIAL (decl))
        {
           print_class (type);
           return;
        }
      }
   }

   tree id = DECL_NAME (decl);
   string name = id ? IDENTIFIER_POINTER (id) : "<unnamed>";

   // see gcc-4.9.2/gcc/cp/error.c and cxx-pretty-print.c

   io.Put (decl_as_string (decl,
                           TFF_UNQUALIFIED_NAME |
                           TFF_TEMPLATE_HEADER |
                           // TFF_DECL_SPECIFIERS | // problem with Qt5
                           TFF_RETURN_TYPE |
                           TFF_FUNCTION_DEFAULT_ARGUMENTS |
                           TFF_EXCEPTION_SPECIFICATION
                           ));
   io.PutLn (";");

   if (false)
   {
      io.Put (string ("// ") /* + decl_scope (decl) + "::" */ + name);

      io.Put (string (" // ") + tree_name (dc));

      if (type)
         // io.Put (string (", type ") + tree_name (tc));
         io.Put (decl_as_string (decl, TFF_DECL_SPECIFIERS
                                       |TFF_RETURN_TYPE
                                       |TFF_FUNCTION_DEFAULT_ARGUMENTS
                                       |TFF_EXCEPTION_SPECIFICATION));

      io.PutLn (string (" // ") +
                DECL_SOURCE_FILE (decl) + ":" +
                IntToStr (DECL_SOURCE_LINE (decl)) + ":" +
                IntToStr (DECL_SOURCE_COLUMN (decl)) );
   }
}

/* ---------------------------------------------------------------------- */

void collect (tree ns)
{
   cp_binding_level * level = NAMESPACE_LEVEL (ns);

   // Collect declarations.
   for (tree decl = level->names; decl != NULL; decl = TREE_CHAIN (decl))
   {
       if (! DECL_IS_BUILTIN (decl))
       {
          print_decl (decl);
       }
   }

   // Traverse namespaces.
   for (tree decl = level->namespaces; decl != NULL; decl = TREE_CHAIN (decl))
   {
      if (! DECL_IS_BUILTIN (decl))
      {
         string name = decl == global_namespace ? "" : decl== NULL ? IDENTIFIER_POINTER (decl) : "<unnamed>";
         io.PutLn ("namespace " + name);
         io.PutLn ("{");
         io.IncIndent ();

         collect (decl);

         io.DecIndent ();
         io.PutLn ("}");
      }
   }
}

/* ---------------------------------------------------------------------- */

string output_name = "";

extern "C" void gate_callback (void*, void*)
{
   // If there were errors during compilation,
   // let GCC handle the exit.
   if (errorcount || sorrycount)
      return;

   int result = 0;

   cerr << "processing " << main_input_filename << endl;

   io.Open (output_name);

   set_error_handlers (); // !?
   collect (global_namespace);

   io.WriteText ();

   exit (result);
}

/* ---------------------------------------------------------------------- */

extern "C" int plugin_init (plugin_name_args * info, plugin_gcc_version * version)
{
   int result = 0;

   cerr << "starting " << info->base_name << endl;

   for (int i = 0; i < info->argc; i++)
   {
      string key = info->argv[i].key; // convert to std::string, important for compare
      string value = info->argv[i].value;
      cerr << "parameter " << key << ", " << value << endl;
      if (key == "output")
      {
         output_name = value;
         cerr << "output " << output_name << endl;
      }
   }

   // Disable assembly output.
   asm_file_name = HOST_BIT_BUCKET;

   // Register callbacks.
   register_callback (info->base_name,
                      PLUGIN_OVERRIDE_GATE,
                      &gate_callback,
                      0);

   return result;
}

/* ---------------------------------------------------------------------- */

// kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
